import errno
import os

import genmsg
import genmsg.msgs
from genmsg import MsgContext, MsgSpec, SrvSpec, gentools
from genpy.generator import compute_outfile_name, make_python_safe

from ._typing import TYPE_CHECKING
from .converter import convert_genpy_init, convert_message_class, convert_service_class
from .stub_element import ClassElement, EmptyLinesElement, ImportsElement, ModuleElement

if TYPE_CHECKING:
    from typing import Callable, Dict, Iterator, List, TypeVar

    SpecType = TypeVar("SpecType", MsgSpec, SrvSpec)
    LoaderType = Callable[[MsgContext, str, str], SpecType]
    GeneratorType = Callable[[str, SpecType], Iterator[str]]
    ModuleFinderType = Callable[[str], Iterator[str]]


GENMSG_EXT_LENGTH = len(genmsg.EXT_MSG)


def generate_message_stub(package, spec):
    # type: (str, MsgSpec) -> Iterator[str]
    pyspec = make_python_safe(spec)
    imports = ImportsElement()
    genclass = convert_message_class(package, pyspec, imports)

    module = ModuleElement()
    module.add_element(imports)
    module.add_element(EmptyLinesElement())
    module.add_element(genclass)

    for line in module.generate():
        yield line


def generate_service_stub(package, spec):
    # type: (str, SrvSpec) -> Iterator[str]
    imports = ImportsElement()
    message_classes = []  # type: List[ClassElement]

    for message_spec in (spec.request, spec.response):
        pyspec = make_python_safe(message_spec)
        message_classes.append(convert_message_class(package, pyspec, imports))

    module = ModuleElement()
    module.add_element(imports)
    module.add_element(EmptyLinesElement())
    module.add_element(message_classes[0])  # request class
    module.add_element(EmptyLinesElement())
    module.add_element(message_classes[1])  # response class
    module.add_element(EmptyLinesElement())
    module.add_element(convert_service_class(spec))  # service class

    for line in module.generate():
        yield line


def _make_dirs(path):
    # type: (str) -> None
    # NOTE: This script will be executed in parallel.
    # So atomic operations are needed to avoid a race condition
    # e.g. Calling `makedir` after `exist` check might cause a race condition
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def _compute_outfile_path(outdir, target_path):
    # type: (str, str) -> str
    outpath = compute_outfile_name(
        outdir, os.path.basename(target_path), genmsg.EXT_MSG
    )  # type: str

    # Replace suffix .py with .pyi
    assert outpath.endswith(".py")
    outpath = outpath + "i"
    return outpath


def _load_spec(
    msg_context,  # type: MsgContext
    loader,  # type: LoaderType[SpecType]
    package,  # type: str
    target_path,  # type: str
    search_paths,  # type: Dict[str, List[str]]
):
    # type: (...) -> SpecType
    filename = os.path.basename(target_path)
    full_type = gentools.compute_full_type_name(package, filename)
    spec = loader(msg_context, target_path, full_type)

    # NOTE: We don't need to load dependency as we just emit the type information
    # The following lines are intended to just verify the message.
    try:
        genmsg.msg_loader.load_depends(msg_context, spec, search_paths)
    except genmsg.InvalidMsgSpec as e:
        raise RuntimeError("Failed to load message: {}".format(e))

    return spec


def _find_all_genpy_modules_genmsg(package_dir):
    # type: (str) -> Iterator[str]
    for path in os.listdir(package_dir):
        if not os.path.isfile(os.path.join(package_dir, path)):
            continue

        if path.endswith(genmsg.EXT_MSG) or path.endswith(genmsg.EXT_SRV):
            yield path[:-GENMSG_EXT_LENGTH]


def _find_all_genpy_modules_py(package_dir):
    # type: (str) -> Iterator[str]
    for path in os.listdir(package_dir):
        if path == "__init__.py" or not os.path.isfile(os.path.join(package_dir, path)):
            continue

        if path.startswith("_") and path.endswith(".py"):
            yield path[1:-3]


def generate_stub(
    msg_context,  # type: MsgContext
    loader,  # type: LoaderType[SpecType]
    generator,  # type: GeneratorType[SpecType]
    package,  # type: str
    target_path,  # type: str
    outdir,  # type: str
    search_paths,  # type: Dict[str, List[str]]
):
    # type: (...) -> str
    spec = _load_spec(msg_context, loader, package, target_path, search_paths)
    outpath = _compute_outfile_path(outdir, target_path)

    _make_dirs(outdir)
    with open(outpath, "w") as f:
        for line in generator(package, spec):
            f.write("{}\n".format(line))

    return outpath


GenPyModuleFinders = {
    "py": _find_all_genpy_modules_py,
    "genmsg": _find_all_genpy_modules_genmsg,
}  # type: Dict[str, ModuleFinderType]


def generate_module_stub(package_dir, outdir, module_finder):
    # type: (str, str, str) -> str
    module_finder_impl = GenPyModuleFinders.get(module_finder)
    if module_finder_impl is None:
        raise ValueError("Unknown module finder: {}".format(module_finder))
    genmsg_modules = sorted(module_finder_impl(package_dir))
    imports = convert_genpy_init(genmsg_modules)

    _make_dirs(outdir)
    outpath = os.path.join(outdir, "__init__.pyi")
    with open(outpath, "w") as f:
        for line in imports.generate(0):
            f.write("{}\n".format(line))

    return outpath



import functools

from ._typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, TypeVar

    TResult = TypeVar("TResult")

try:
    from functools import lru_cache  # type: ignore
except ImportError:
    # Provide a simple cache for Python2
    def lru_cache():
        # type: (...) -> Callable[..., Callable[..., TResult]]
        def _lru_cache(func):
            # type: (Callable[..., TResult]) -> Callable[..., TResult]
            cache = {}  # type: Dict[Any, Any]

            @functools.wraps(func)
            def wrapper(*args):
                # type: (Any) -> TResult
                key = tuple(args)
                if key not in cache:
                    ret = func(*args)
                    cache[key] = ret
                else:
                    ret = cache[key]

                return ret

            return wrapper

        return _lru_cache



import genmsg
import genmsg.msgs
from genmsg import MsgSpec, SrvSpec
from genpy.generator import compute_pkg_type

from ._compat import lru_cache
from ._typing import TYPE_CHECKING
from .stub_element import (
    ClassElement,
    ClassMethodElement,
    CommentElement,
    EmptyLinesElement,
    FieldElement,
    ImportsElement,
    ParameterElement,
)

if TYPE_CHECKING:
    from typing import Dict, List, Sequence, Tuple


_GENMSG_PRIMITIVES = {
    "byte": "byte",
    "char": "str",
    "int8": "int",
    "int16": "int",
    "int32": "int",
    "int64": "int",
    "uint8": "int",
    "uint16": "int",
    "uint32": "int",
    "uint64": "int",
    "float32": "float",
    "float64": "float",
    "bool": "bool",
    "string": "str",
}

_GENPY_DEFINED = {
    genmsg.HEADER: ("std_msgs.msg._Header.Header", "std_msgs.msg._Header"),
    genmsg.TIME: ("genpy.Time", "genpy"),
    genmsg.DURATION: ("genpy.Duration", "genpy"),
}  # type: Dict[str, Tuple[str, str]]


@lru_cache()
def _get_genmsg_type(first_party_package, field_type, imports):
    # type: (str, str, ImportsElement) -> str

    """Get the corresponding python type for the given ROS type.

    .. warning::
        As this method is decorated by lru_cache, make sure that
        you are using another instance of ImportsElement when
        calling this method in a different context.
        (e.g. creating a different module)
    """
    primitive = _GENMSG_PRIMITIVES.get(field_type)
    if primitive is not None:
        return primitive

    if field_type.endswith("]"):
        base_field_type, is_array, array_len = genmsg.msgs.parse_type(field_type)
        assert is_array
        # TODO: Handle array_len (use typing.Annotated?)
        base_type = _get_genmsg_type(first_party_package, base_field_type, imports)
        return "typing.List[{}]".format(base_type)

    special = _GENPY_DEFINED.get(field_type)
    if special is not None:
        full_name, module = special
        imports.add_third_party_module(module, None)
        return full_name

    package, type_ = compute_pkg_type(first_party_package, field_type)
    module = "{}.msg".format(package)
    imports.add_third_party_module(module, None)

    return "{}.{}".format(module, type_)


def _convert_message_fields(first_party_package, spec, imports):
    # type: (str, MsgSpec, ImportsElement) -> List[FieldElement]
    fields = []  # type: List[FieldElement]
    for field in spec.fields():
        type_, name = field
        fields.append(
            FieldElement(
                name,
                _get_genmsg_type(first_party_package, type_, imports),
            )
        )

    return fields


def _convert_message_constants(first_party_package, spec, imports):
    # type: (str, MsgSpec, ImportsElement) -> List[FieldElement]
    constants = []  # type: List[FieldElement]
    for constant in spec.constants:
        constants.append(
            FieldElement(
                constant.name,
                _get_genmsg_type(first_party_package, constant.type, imports),
            )
        )

    return constants


def convert_message_class(first_party_package, spec, imports):
    # type: (str, MsgSpec, ImportsElement) -> ClassElement
    imports.add_third_party_module("genpy", None)
    imports.add_system_module("typing", None)
    imports.add_system_module("types", None)

    msgclass = ClassElement(spec.short_name, "genpy.Message")

    # Add private fields
    msgclass.add(FieldElement("_md5sum", "str"))
    msgclass.add(FieldElement("_type", "str"))
    msgclass.add(FieldElement("_has_header", "bool"))
    msgclass.add(FieldElement("_full_text", "str"))
    msgclass.add(FieldElement("__slots__", "typing.List[str]"))
    msgclass.add(FieldElement("_slot_types", "typing.List[str]"))

    message_constants = _convert_message_constants(first_party_package, spec, imports)
    if len(message_constants) > 0:
        msgclass.add(EmptyLinesElement())
        msgclass.add(CommentElement("Constants"))
        for c in message_constants:
            msgclass.add(c)

    message_fields = _convert_message_fields(first_party_package, spec, imports)
    if len(message_fields) > 0:
        msgclass.add(EmptyLinesElement())
        msgclass.add(CommentElement("Fields"))
        for f in message_fields:
            msgclass.add(f)

    msgclass.add(EmptyLinesElement())

    # Add __init__ method that accepts message fields as the parameter
    init_method = ClassMethodElement("__init__", "None")
    for field in message_fields:
        init_method.add_parameter(
            ParameterElement(field.name, field.type, has_default=True)
        )

    # NOTE: Emit *args and **kwds in order to align with the actual implementation
    init_method.add_parameter(ParameterElement("*args", "typing.Any"))
    init_method.add_parameter(ParameterElement("**kwds", "typing.Any"))

    msgclass.add(init_method)

    # Add private methods except pattern methods like `_get_struct_I`
    msgclass.add(ClassMethodElement("_get_types", "typing.List[str]"))

    # Add public methods
    msgclass.add(
        ClassMethodElement(
            "serialize", "None", [ParameterElement("buff", "typing.StringIO")]
        )
    )
    msgclass.add(
        ClassMethodElement(
            "deserialize", spec.short_name, [ParameterElement("str", "str")]
        )
    )
    msgclass.add(
        ClassMethodElement(
            "serialize_numpy",
            "None",
            [
                ParameterElement("buff", "typing.StringIO"),
                ParameterElement("numpy", "types.ModuleType"),
            ],
        )
    )
    msgclass.add(
        ClassMethodElement(
            "deserialize_numpy",
            spec.short_name,
            [
                ParameterElement("str", "str"),
                ParameterElement("numpy", "types.ModuleType"),
            ],
        )
    )

    return msgclass


def convert_service_class(spec):
    # type: (SrvSpec) -> ClassElement
    srvclass = ClassElement(spec.short_name, "object")
    srvclass.add(FieldElement("_type", "str"))
    srvclass.add(FieldElement("_md5sum", "str"))
    srvclass.add(FieldElement("_request_class", "str"))
    srvclass.add(FieldElement("_response_class", "str"))

    return srvclass


def convert_genpy_init(modules):
    # type: (Sequence[str]) -> ImportsElement
    imports = ImportsElement()
    for m in modules:
        imports.add_third_party_module("._{}".format(m), "*")

    return imports



import argparse
import os

from genmsg import MsgContext, command_line, msg_loader

from . import generator
from ._typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Dict, List, Optional

    SrvMsgCallback = Callable[
        [str, List[str], Optional[str], Dict[str, List[str]]], None
    ]


def run_message(
    package,  # type: str
    package_files,  # type: List[str]
    outdir,  # type: Optional[str]
    search_paths,  # type: Dict[str, List[str]]
):
    # type: (...) -> None
    msg_context = MsgContext.create_default()

    for target_path in package_files:
        target_path = os.path.abspath(target_path)
        output_dir = outdir
        if output_dir is None:
            output_dir = os.path.abspath(os.path.join(target_path, os.path.pardir))

        generator.generate_stub(
            msg_context,
            msg_loader.load_msg_from_file,
            generator.generate_message_stub,
            package,
            target_path,
            output_dir,
            search_paths,
        )


def run_service(
    package,  # type: str
    package_files,  # type: List[str]
    outdir,  # type: Optional[str]
    search_paths,  # type: Dict[str, List[str]]
):
    # type: (...) -> None
    msg_context = MsgContext.create_default()

    for target_path in package_files:
        target_path = os.path.abspath(target_path)
        output_dir = outdir
        if output_dir is None:
            output_dir = os.path.abspath(os.path.join(target_path, os.path.pardir))

        generator.generate_stub(
            msg_context,
            msg_loader.load_srv_from_file,
            generator.generate_service_stub,
            package,
            target_path,
            output_dir,
            search_paths,
        )


def run_module(
    package_dir,  # type: str
    outdir,  # type: str
    module_finder,  # type: str
):
    # type: (...) -> None
    generator.generate_module_stub(package_dir, outdir, module_finder)


def _start_msg_srv(args):
    # type: (argparse.Namespace) -> None
    search_paths = command_line.includepath_to_dict(
        args.include_path
    )  # type: Dict[str, List[str]]
    args.handler(args.package, args.files, args.out_dir, search_paths)


def _start_module(args):
    # type: (argparse.Namespace) -> None
    package_dir = args.package_dir
    out_dir = args.out_dir
    if out_dir is None:
        out_dir = package_dir

    run_module(package_dir, out_dir, args.module_finder)


def _setup_module_options(parser):
    # type: (argparse.ArgumentParser) -> None
    parser.add_argument(
        "package_dir",
        type=str,
        help="Package directory to create __init__.pyi",
    )
    parser.add_argument(
        "--module-finder",
        type=str,
        choices=generator.GenPyModuleFinders.keys(),
        help=(
            "Method to find generated messages/services. "
            "(genmsg: Files with .msg, .srv would be regarded as modules. "
            "py: Python files starting with `_` would be regarded as modules.)"
        ),
        default="genmsg",
    )
    parser.add_argument(
        "--out-dir",
        type=str,
        help=(
            "Output directory. "
            "If the option is unset, __init__.pyi will be generated in the same "
            "directory as package_dir."
        ),
    )
    parser.set_defaults(func=_start_module)


def _setup_msg_srv_options(parser, handler):
    # type: (argparse.ArgumentParser, SrvMsgCallback) -> None
    parser.add_argument(
        "package", type=str, help="Package name of given files", default="out"
    )
    parser.add_argument("files", type=str, help="Files to generate stubs", nargs="+")
    parser.add_argument(
        "--out-dir",
        type=str,
        help=(
            "Output directory. "
            "If the option is unset, each stub file will be generated in the same "
            "directory as each input."
        ),
    )
    parser.add_argument(
        "--include-path",
        "-I",
        type=str,
        action="append",
        help="Include paths for processing given files",
    )
    parser.set_defaults(func=_start_msg_srv)
    parser.set_defaults(handler=handler)


def cli():
    # type: (...) -> None
    prog = "genmypy"
    parser = argparse.ArgumentParser(
        prog,
        description="""Generate python stub files from genmsg specs

Examples:
$ {0} msg custom_msgs custom_msgs/msg/Custom.msg
$ {0} msg sensor_msgs --out-dir out \\
    -Istd_msgs:/opt/ros/melodic/share/std_msgs/msg \\
    -Isensor_msgs:/opt/ros/melodic/share/sensor_msgs/msg \\
    /opt/ros/melodic/share/sensor_msgs/msg/PointCloud2.msg
$ {0} srv custom_msgs custom_msgs/srv/Custom.msg
$ {0} srv nav_msgs --out-dir out \\
    -Istd_msgs:/opt/ros/melodic/share/std_msgs/msg \\
    -Isensor_msgs:/opt/ros/melodic/share/sensor_msgs/msg \\
    /opt/ros/melodic/share/sensor_msgs/srv/SetCameraInfo.srv
$ {0} module custom_msgs/msg/""".format(
            prog
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparser = parser.add_subparsers()
    _setup_msg_srv_options(
        subparser.add_parser("msg", help="Generate stub files from .msg files"),
        run_message,
    )
    _setup_msg_srv_options(
        subparser.add_parser("srv", help="Generate stub files from .srv files"),
        run_service,
    )
    _setup_module_options(
        subparser.add_parser(
            "module", help="Generate __init__.pyi from a msg/srv directory"
        )
    )
    args = parser.parse_args()
    args.func(args)



from setuptools import find_packages, setup

setup(
    name="genmypy",
    version="0.3.0",
    packages=find_packages(exclude=["tests"]),
    description="A Python stub generator from genmsg specs",
    long_description=open("README.rst").read(),
    author="Yuki Igarashi, Tamamki Nishino",
    author_email="me@bonprosoft.com, otamachan@gmail.com",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python",
    ],
    license="Apache License 2.0",
    url="https://github.com/rospypi/genmypy",
    install_requires=[
        "genmsg",
        "genpy",
    ],
    extras_require={
        "dev": [
            # NOTE: pytest>=5.0 doesn't support py2
            "pytest>=4.6,<5.0",
            "typing; python_version=='2.7'",
        ],
        # NOTE: We don't support python2 for the lint environment
        "lint": [
            "black==20.8b1",
            "flake8-bugbear==21.4.3",
            "flake8==3.9.1",
            "isort==5.1.4",
            "mypy==0.790",
            "pysen>=0.9,<0.10",
        ],
    },
    package_data={"genmypy": ["py.typed"]},
    scripts=["scripts/genmypy"],
)



import collections

from ._typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import DefaultDict, Iterator, List, Optional, Sequence, Set, Tuple

INDENT_SIZE = 4
FUNCTION_PARAM_MULTILINE_THRESHOLD = 70


def _indent(depth):
    # type: (int) -> str
    return " " * INDENT_SIZE * depth


def _format(content, depth):
    # type: (str, int) -> str
    return "{}{}".format(_indent(depth), content)


class ModuleElement:
    def __init__(self):
        # type: () -> None
        self.statements = []  # type: List[StatementElementBase]

    def add_element(self, statement):
        # type: (StatementElementBase) -> None
        self.statements.append(statement)

    def generate(self):
        # type: () -> Iterator[str]
        for statement in self.statements:
            for line in statement.generate(0):
                yield line


class StatementElementBase(object):
    def generate(self, indent):
        # type: (int) -> Iterator[str]
        raise NotImplementedError()


def _generate_imports(items, indent):
    # type: (DefaultDict[str, Set[Optional[str]]], int) -> Tuple[List[str], List[str]]
    imports = []  # type: List[str]
    from_imports = []  # type: List[str]
    for module, elements in sorted(items.items()):
        if None in elements:
            imports.append(_format("import {}".format(module), indent))

        names = [e for e in elements if e is not None]
        if len(names) == 0:
            continue

        joined = ", ".join(sorted(names))
        from_imports.append(_format("from {} import {}".format(module, joined), indent))

    return imports, from_imports


class ImportsElement(StatementElementBase):
    def __init__(self):
        # type: () -> None
        self._system_modules = collections.defaultdict(
            set
        )  # type: DefaultDict[str, Set[Optional[str]]]
        self._third_party_modules = collections.defaultdict(
            set
        )  # type: DefaultDict[str, Set[Optional[str]]]

    def add_system_module(self, module, name):
        # type: (str, Optional[str]) -> None
        self._system_modules[module].add(name)

    def add_third_party_module(self, module, name):
        # type: (str, Optional[str]) -> None
        self._third_party_modules[module].add(name)

    def generate(self, indent):
        # type: (int) -> Iterator[str]
        imports, from_imports = _generate_imports(self._system_modules, indent)
        for line in imports:
            yield line
        for line in from_imports:
            yield line

        if len(self._system_modules) > 0:
            yield ""

        imports, from_imports = _generate_imports(self._third_party_modules, indent)
        for line in imports:
            yield line
        for line in from_imports:
            yield line


class EmptyLinesElement(StatementElementBase):
    def __init__(self, lines=1):
        # type: (int) -> None
        self.lines = lines

    def generate(self, indent):
        # type: (int) -> Iterator[str]
        # NOTE: Do not use _format(...) for empty lines
        for _ in range(self.lines):
            yield ""


class CommentElement(StatementElementBase):
    def __init__(self, comment):
        # type: (str) -> None
        self.comment = comment

    def generate(self, indent):
        # type: (int) -> Iterator[str]
        for line in self.comment.splitlines():
            yield _format("# {}".format(line), indent)


class ClassElement(StatementElementBase):
    def __init__(self, name, base=None):
        # type: (str, Optional[str]) -> None
        self.name = name
        self.base = base
        self.members = []  # type: List[StatementElementBase]

    def add(self, member):
        # type: (StatementElementBase) -> None
        self.members.append(member)

    def generate(self, indent):
        # type: (int) -> Iterator[str]
        if self.base is not None:
            yield _format("class {}({}):".format(self.name, self.base), indent)
        else:
            yield _format("class {}:".format(self.name), indent)

        for m in self.members:
            for line in m.generate(indent + 1):
                yield line


class FieldElement(StatementElementBase):
    def __init__(self, name, type):
        # type: (str, str) -> None
        self.name = name
        self.type = type

    def generate(self, indent):
        # type: (int) -> Iterator[str]
        yield _format("{}: {}".format(self.name, self.type), indent)


class FunctionElement(StatementElementBase):
    def __init__(self, name, return_type, params=None):
        # type: (str, str, Optional[Sequence[ParameterElement]]) -> None
        self.name = name
        self.return_type = return_type
        self.params = []  # type: List[ParameterElement]

        if params is not None:
            self.params.extend(params)

    def add_parameter(self, param):
        # type: (ParameterElement) -> None
        self.params.append(param)

    def generate(self, indent):
        # type: (int) -> Iterator[str]
        param_expression = ""

        params = [p.generate_expression() for p in self.params]
        total_length = sum(len(e) for e in params)
        if total_length > FUNCTION_PARAM_MULTILINE_THRESHOLD:
            param_defs = ",\n".join(_format(p, indent + 1) for p in params)
            param_expression = "\n{},\n{}".format(param_defs, _format("", indent))
        else:
            param_expression = ", ".join(params)

        generated = _format(
            "def {}({}) -> {}: ...".format(
                self.name,
                param_expression,
                self.return_type,
            ),
            indent,
        )
        for line in generated.splitlines():
            yield line


class ClassMethodElement(FunctionElement):
    def __init__(self, name, return_type, params=None):
        # type: (str, str, Optional[Sequence[ParameterElement]]) -> None
        super(ClassMethodElement, self).__init__(name, return_type)
        self.add_parameter(ParameterElement("self", None))
        if params is not None:
            for p in params:
                self.add_parameter(p)


class ParameterElement:
    def __init__(self, name, type, has_default=False):
        # type: (str, Optional[str], bool) -> None
        self.name = name
        self.type = type
        self.has_default = has_default

    def generate_expression(self):
        # type: () -> str
        ret = str(self.name)
        if self.type is not None:
            ret = "{}: {}".format(ret, self.type)

        if self.has_default:
            ret = "{} = ...".format(ret)

        return ret
